/* DVB USB compliant Linux driver for the Friio USB2.0 ISDB-T receiver.
 *
 * Copyright (C) 2009 Akihiro Tsukada <tskd2@yahoo.co.jp>
 *
 * This module is based off the the gl861 and vp702x modules.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, version 2.
 *
 * see Documentation/dvb/README.dvb-usb for more information
 */
#include <linux/init.h>
#include <linux/string.h>
#include <linux/slab.h>

#include "friio.h"

/* frontend (demod) private data */
struct jdvbt90502_state {
	struct i2c_adapter *i2c;
	struct dvb_frontend frontend;
	struct jdvbt90502_config config;
	struct usb_device *udev;
};


/* helper functions */
static int do_reg_read(struct jdvbt90502_state *state, const u8 addr,
			const u8 reg, u8 *buf, const size_t count)
{
	int ret;
	u8 wbuf[2];
	struct i2c_msg msg[2];

	wbuf[0] = reg;
	wbuf[1] = (reg == JDVBT90502_2ND_I2C_REG) ? *buf : 0;

	msg[0].addr = addr;
	msg[0].flags = 0;
	msg[0].buf = wbuf;
	msg[0].len = sizeof(wbuf);

	msg[1].addr = msg[0].addr;
	msg[1].flags = I2C_M_RD;
	msg[1].buf = buf;
	msg[1].len = count;

	ret = i2c_transfer(state->i2c, msg, 2);
	if (ret != 2) {
		deb_fe(" reg read failed.\n");
		return -EREMOTEIO;
	}
	return 0;
}

static int do_reg_write(struct jdvbt90502_state *state, const u8 addr,
			const u8 reg, const u8 val)
{
	struct i2c_msg msg;
	u8 wbuf[2];

	wbuf[0] = reg;
	wbuf[1] = val;

	msg.addr = addr;
	msg.flags = 0;
	msg.buf = wbuf;
	msg.len = sizeof(wbuf);

	if (i2c_transfer(state->i2c, &msg, 1) != 1) {
		deb_fe(" reg write failed.");
		return -EREMOTEIO;
	}
	return 0;
}

/* demodulator I2C access */
/* NOTE: when I2C through mode access is used for accessing the tuner, */
/* buf[0] is used to designate the tuner register */
static int jdvbt90502_reg_read(struct jdvbt90502_state *state,
			       const u8 reg, u8 *buf, const size_t count)
{
	return do_reg_read(state, state->config.demod_address,
				reg, buf, count);
}

static int jdvbs90502_reg_read(struct jdvbt90502_state *state,
				const u8 reg, u8 *buf, const size_t count)
{
	return do_reg_read(state, state->config.bk_demod_address,
				reg, buf, count);
}


static int jdvbt90502_single_reg_write(struct jdvbt90502_state *state,
				       const u8 reg, const u8 val)
{
	return do_reg_write(state, state->config.demod_address, reg, val);
}

static int jdvbs90502_single_reg_write(struct jdvbt90502_state *state,
				       const u8 reg, const u8 val)
{
	return do_reg_write(state, state->config.bk_demod_address, reg, val);
}

static int usb_ctrl_sends(struct jdvbt90502_state *state,uint16_t data[],size_t length,uint8_t *rcvbuf,size_t recv_len){
  uint8_t sbuf[20];
  u16 index;
  u16 value;
  u8 req,rtype;
  int rval;
  u16 dlength;
  int pipe,i,j;
  int line=0;

  for(i=0;i<length;i++){
    rtype	=data[i];
    req	=data[++i];
    value	=data[++i];
    index	=data[++i];
    dlength	=data[++i];
    pipe=(rtype==0xc0)?usb_rcvctrlpipe(state->udev,0):usb_sndctrlpipe(state->udev,0);
    if(dlength>0){
      memset(sbuf,0,sizeof(sbuf));
      if(rtype!=0xc0)for(j=0;j<dlength;j++)sbuf[j]=data[++i];
      rval=usb_control_msg(state->udev,pipe,req,rtype,value,index,sbuf,dlength,1000);
      if(rtype==0xc0 && recv_len>0){
	for(j=0;j<recv_len;j++)rcvbuf[j]=sbuf[j];
      }
    }else{
      rval=usb_control_msg(state->udev,pipe,req,rtype,value,index,sbuf,dlength,2000);
    }
  }
  return 0;
}
  
/* read tuner status byte via the demodulator's I2C register */
/* note: Win box reads a 8-byte block at the I2C addr 0x30 from reg:0x80 */
static int jdvbt90502_pll_read(struct jdvbt90502_state *state, u8 *result)
{
	int ret;

	/* +1 for reading */
	u8 pll_addr_byte = (state->config.pll_address << 1) + 1;

	*result = pll_addr_byte;

	ret = jdvbt90502_reg_read(state, JDVBT90502_2ND_I2C_REG, result, 1);
	if (ret)
		goto error;

	deb_fe("PLL read val:%02x\n", *result);
	return 0;

error:
	deb_fe("%s:ret == %d\n", __func__, ret);
	return -EREMOTEIO;
}

static int jdvbs90502_pll_read(struct jdvbt90502_state *state, u8 *result)
{
	int ret;

	/* +1 for reading */
	u8 pll_addr_byte = (state->config.bk_pll_address << 1) + 1;

	*result = pll_addr_byte;

	ret = jdvbs90502_reg_read(state, JDVBT90502_2ND_I2C_REG, result, 1);
	if (ret)
		goto error;

	deb_fe("PLL read val:%02x\n", *result);
	return 0;

error:
	deb_fe("%s:ret == %d\n", __func__, ret);
	return -EREMOTEIO;
}


/* TMCC preset commands*/
const static u8 demod_reset_cmd[][2] = {
	{0xB2, 0x3D},
	{0xB3, 0x25},
	{0xB4, 0x8B},
	{0xB5, 0x4B},
	{0xB6, 0x3F},
	{0xB7, 0xFF},
	{0xB8, 0xFF},
	{0xB9, 0xFC},
	{0x47, 0x80},
};
const static int reset_cmd_len =
	sizeof(demod_reset_cmd)/sizeof(demod_reset_cmd[0]);

/* set pll frequency via the demodulator's I2C register */
static int jdvbt90502_set_freq(struct jdvbt90502_state *state, u32 freq)
{
	int i;
	int ret;
	int retry;
	u8 res1;

	u8 pll_freq_cmd[PLL_CMD_LEN];
	u8 pll_agc_cmd[PLL_CMD_LEN];
	struct i2c_msg msg[2];
	u32 f;

	deb_fe("%s: freq=%d, step=%d\n", __func__, freq,
	       state->frontend.ops.info.frequency_stepsize);
	/* freq -> oscilator frequency conversion. */
	/* freq: 473,000,000 + n*6,000,000 [+ 142857 (center freq. shift)] */
	f = freq / state->frontend.ops.info.frequency_stepsize;
	/* add 399[1/7 MHZ] = 57MHz for the IF  */
	f += 399;
	/* add center frequency shift if necessary */
	if (f % 7 == 0)
		f++;
	pll_freq_cmd[DEMOD_REDIRECT_REG] = JDVBT90502_2ND_I2C_REG; /* 0xFE */
	pll_freq_cmd[ADDRESS_BYTE] = state->config.pll_address << 1;
	pll_freq_cmd[DIVIDER_BYTE1] = (f >> 8) & 0x7F;
	pll_freq_cmd[DIVIDER_BYTE2] = f & 0xFF;
	pll_freq_cmd[CONTROL_BYTE] = 0xB2; /* ref.divider:28, 4MHz/28=1/7MHz */
	pll_freq_cmd[BANDSWITCH_BYTE] = 0x08;	/* UHF band */

	msg[0].addr = state->config.demod_address;
	msg[0].flags = 0;
	msg[0].buf = pll_freq_cmd;
	msg[0].len = sizeof(pll_freq_cmd);

	ret = i2c_transfer(state->i2c, &msg[0], 1);
	if (ret != 1)
		goto error;

	udelay(50);

	pll_agc_cmd[DEMOD_REDIRECT_REG] = pll_freq_cmd[DEMOD_REDIRECT_REG];
	pll_agc_cmd[ADDRESS_BYTE] = pll_freq_cmd[ADDRESS_BYTE];
	pll_agc_cmd[DIVIDER_BYTE1] = pll_freq_cmd[DIVIDER_BYTE1];
	pll_agc_cmd[DIVIDER_BYTE2] = pll_freq_cmd[DIVIDER_BYTE2];
	pll_agc_cmd[CONTROL_BYTE] = 0x9A; /*  AGC_CTRL instead of BANDSWITCH */
	pll_agc_cmd[AGC_CTRL_BYTE] = 0x50;
	/* AGC Time Constant 2s, AGC take-over point:103dBuV(lowest) */

	msg[1].addr = msg[0].addr;
	msg[1].flags = 0;
	msg[1].buf = pll_agc_cmd;
	msg[1].len = sizeof(pll_agc_cmd);

	ret = i2c_transfer(state->i2c, &msg[1], 1);
	if (ret != 1)
		goto error;

	/* wait for the PLL to be locked */
#define RETRY_COUNT 5
	for (retry = 0; retry < RETRY_COUNT; retry++) {
		mdelay(1);
		ret = jdvbt90502_pll_read(state, &res1);
		if (ret)
			goto error;
		if (res1 & PLL_STATUS_LOCKED)
			break;
	}
	if (retry == RETRY_COUNT) {
		deb_fe("%s:failed to lock PLL.\n", __func__);
		return -EREMOTEIO;
	}

	/* preset TMCC */
	for (i = 0; i < reset_cmd_len; i++) {
		ret = jdvbt90502_single_reg_write(state,
						demod_reset_cmd[i][0],
						demod_reset_cmd[i][1]);
		if (ret)
			goto error;
	}

	/* I don't know what these cmds are for,  */
	/* but the USB log on a windows box contains them */
	ret = jdvbt90502_single_reg_write(state, 0x01, 0x40);
	if (ret)
		goto error;

	return 0;
error:
	deb_fe("%s:ret == %d\n", __func__, ret);
	return -EREMOTEIO;
}

static int jdvbs90502_set_freq(struct jdvbt90502_state *state,
				u32 freq, u16 tsid)
{
	int ret;
	int retry;
	u8 result;
	u16 f0 = freq * 8 / 4000;	/* 8:ref divider, 4000[kHz]: XT */
	u8 freq_cmd[] = {
		JDVBT90502_2ND_I2C_REG,
		state->config.bk_pll_address << 1, 0x00,
		(f0 >> 8), (f0 & 0xff)
	};
	u8 tsid_cmd[] = {0x8F, (tsid >> 8), (tsid & 0xff)};
	struct i2c_msg msg;

	deb_fe("%s: freq=%d\n", __func__, freq);

	/* set frequency */
	msg.addr = state->config.bk_demod_address;
	msg.flags = 0;
	msg.len = sizeof(freq_cmd);
	msg.buf = freq_cmd;
	ret = i2c_transfer(state->i2c, &msg, 1);
	if (ret != 1)
		goto error;

	/* wait for the PLL to be locked */
#define RETRY_COUNT_BK_PLL 5
	for (retry = 0; retry < RETRY_COUNT_BK_PLL; retry++) {
		mdelay(1);
		/* FIXME: should be put into ce5037_init()? */
		result = 0x00;
		ret = jdvbs90502_pll_read(state, &result);
		if (ret != 0)
			goto error;
		if (!(result & 0x80))
			continue;

		result = 0x08;
		ret = jdvbs90502_pll_read(state, &result);
		if (ret != 0)
			goto error;
		if (result & 0x80)
			break;
	}

	if (retry == RETRY_COUNT_BK_PLL) {
		deb_fe("%s:failed to lock PLL.\n", __func__);
		return -EREMOTEIO;
	}


	/* reset demodulation */
	ret = jdvbs90502_single_reg_write(state, 0x03, 0x01);
	if (ret)
		goto error;

	/* wait for TMCC sync */
#define RETRY_COUNT_BK_TMCC 5
	for (retry = 0; retry < RETRY_COUNT_BK_TMCC; retry++) {
		mdelay(200);
		ret = jdvbs90502_reg_read(state, 0xC3, &result, 1);
		if (ret != 0)
			goto error;
		if (result & 0x10)
			break;
	}
	if (retry == RETRY_COUNT_BK_TMCC) {
		deb_fe("%s:failed to sync TMCC.\n", __func__);
		return -EREMOTEIO;
	}

	/* set TS-ID */
	msg.addr = state->config.bk_demod_address;
	msg.flags = 0;
	msg.len = sizeof(tsid_cmd);
	msg.buf = tsid_cmd;
	ret = i2c_transfer(state->i2c, &msg, 1);
	if (ret != 1)
		goto error;


	return 0;
error:
	deb_fe("%s:ret == %d\n", __func__, ret);
	return -EREMOTEIO;
}
static uint16_t express_usbdata[] = {
  /* requesttype, request, value, index, length, data.... */
  0x40,0x01,0x0002,0x0011,0x0000,
  0x40,0x01,0x0000,0x0011,0x0000,
  0x40,0x03,0x1200,0x0100,0x0002,0x03,0x80,
  0xc0,0x02,0x1200,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x80,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x01,0x0004,0x0030,0x0000,
  0x40,0x01,0x0001,0x0000,0x0000,
  0x40,0x01,0x00ff,0x0006,0x0000,
  0x40,0x01,0x000f,0x0006,0x0000,
  0x40,0x01,0x0008,0x0033,0x0000,
  0x40,0x01,0x0040,0x0037,0x0000,
  0x40,0x01,0x001f,0x003a,0x0000,
  0x40,0x01,0x00ff,0x003b,0x0000,
  0x40,0x01,0x001f,0x003c,0x0000,
  0x40,0x01,0x00ff,0x003d,0x0000,
  0x40,0x01,0x0000,0x0038,0x0000,
  0x40,0x01,0x0000,0x0035,0x0000,
  0x40,0x01,0x0000,0x0039,0x0000,
  0x40,0x01,0x0000,0x0036,0x0000,
  0x40,0x03,0x1200,0x0100,0x0002,0x03,0x80,
  0xc0,0x02,0x1200,0x0100,0x0002,
  0x40,0x03,0x1200,0x0100,0x0002,0x03,0x81,
  0xc0,0x02,0x1200,0x0100,0x0002,
  0x40,0x03,0x1200,0x0100,0x0004,0x05,0x06,0x05,0x00,
  0x40,0x03,0x1200,0x0100,0x0004,0x05,0x05,0x02,0x42,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x05,0x02,0x40,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x06,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x84,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x00,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x06,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x07,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x08,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x84,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x00,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8c,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8b,0x30,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8c,0x00,0x7a,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x02,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x03,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x04,0xff,0xfc,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x82,0x00,0x0c,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x00,0x00,0x4d,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x01,0x00,0x0c,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x02,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x03,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x04,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x82,0x00,0x04,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x00,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x01,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8a,0x2d,0x98,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x85,0x01,0x14,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x86,0x05,0x0f,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x86,0x05,0x07,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x88,0x1e,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x17,0x03,0x93,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x18,0x87,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1c,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1a,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x9a,0xd2,0x57,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x74,0x00,0x07,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x05,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x06,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x08,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x0d,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x2b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x2b,0x00,0x22,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x2c,0x07,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x06,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x0b,0x02,0x05,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x28,0x07,0x55,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x07,0x00,0x04,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x08,0x04,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x09,0x08,0x14,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0b,0x00,0x1b,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0c,0x77,0x40,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0d,0x00,0x5b,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0e,0x8d,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0f,0x01,0xc9,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x10,0xc3,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x11,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x12,0x00,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x13,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x14,0x00,0x90,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x15,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x16,0xd4,0xc0,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x50,0x1c,0x52,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x51,0x1c,0x46,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x52,0x1c,0x60,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x53,0x1c,0x8d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x54,0x1c,0xaa,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x55,0x1c,0xd6,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x56,0x1d,0x03,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x57,0x1d,0x20,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x58,0x1d,0x4c,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x59,0x1d,0x78,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x5a,0x1d,0x96,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb3,0x20,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb5,0x28,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb6,0x28,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb7,0x28,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb8,0x28,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb9,0x28,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xba,0x28,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xc1,0x06,0x66,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xc2,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xcd,0x20,0x0f,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xce,0x00,0x0f,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd7,0x02,0x3d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd8,0x00,0xa4,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd9,0x00,0xa4,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xda,0x7f,0xf0,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xdb,0x3c,0xcc,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xe6,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x07,0x08,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0c,0x04,0x27,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0e,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x1d,0x00,0x20,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x2b,0x00,0x62,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x52,0x00,0x61,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x87,0x00,0x02,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x07,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x87,0x00,0x31,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x05,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x05,0x02,0x40,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1d,0x01,0x15,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1e,0x66,0xab,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8c,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8c,0x00,0x7b,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8b,0x30,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8c,0x00,0x79,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x9b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x9c,0x03,0x08,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x9b,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x9b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8c,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8b,0x30,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8c,0x00,0x7b,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x84,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x00,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x06,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x07,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x08,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x84,0x00,0x03,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x00,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0xc0,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x05,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x05,0x03,0xfe,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x06,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x06,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x05,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x05,0x03,0xde,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x06,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x06,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x05,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x05,0x03,0xde,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x06,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x06,0x00,0x20,
  0xc0,0x02,0xc000,0x0022,0x0002,
  0xc0,0x02,0xc000,0x001f,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x02,0x00,0x08,
  0x40,0x03,0xc000,0x0100,0x0003,0x03,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x04,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x05,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x06,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x07,0x00,0x02,
  0x40,0x03,0xc000,0x0100,0x0003,0x08,0x01,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x0d,0x0d,0x80,
  0x40,0x03,0xc000,0x0100,0x0003,0x0e,0x00,0x01,
  0x40,0x03,0xc000,0x0100,0x0003,0x0f,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x11,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x12,0x01,0x03,
  0x40,0x03,0xc000,0x0100,0x0003,0x13,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x16,0x00,0x44,
  0x40,0x03,0xc000,0x0100,0x0003,0x17,0x00,0x30,
  0x40,0x03,0xc000,0x0100,0x0003,0x18,0x07,0xff,
  0x40,0x03,0xc000,0x0100,0x0003,0x1b,0x41,0x12,
  0x40,0x03,0xc000,0x0100,0x0003,0x1c,0xff,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x1d,0xc0,0x7f,
  0x40,0x03,0xc000,0x0100,0x0003,0x1e,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x1f,0x00,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x20,0x48,0x7b,
  0x40,0x03,0xc000,0x0100,0x0003,0x10,0x00,0x02,
  0x40,0x03,0xc000,0x0100,0x0003,0x1f,0x01,0x20,
  0x40,0x03,0xc000,0x0100,0x0003,0x15,0x56,0xa4,
  0x40,0x03,0xc000,0x0100,0x0003,0x01,0x6c,0xc8,
  0x40,0x03,0xc000,0x0100,0x0003,0x0f,0x6d,0x81,
  0x40,0x03,0xc000,0x0100,0x0003,0x20,0x00,0x7b,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x20,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0x01,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x84,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x00,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x06,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x07,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x08,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x84,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x00,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8c,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8c,0x00,0x7b,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8b,0x30,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8c,0x00,0x79,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x05,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x05,0x03,0xde,
  0x40,0x03,0x9000,0x0100,0x0002,0x04,0x06,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x06,0x00,0x20,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0xc0,0x00,
  0xc0,0x02,0xc000,0x0002,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x02,0xc0,0x08,
  0x40,0x03,0xc000,0x0100,0x0003,0x17,0x00,0x30,
  0x40,0x03,0xc000,0x0100,0x0003,0x11,0x00,0x5b,
  0x40,0x03,0xc000,0x0100,0x0003,0x12,0xff,0x01,
  0x40,0x03,0xc000,0x0100,0x0003,0x13,0xda,0x93,
  0x40,0x03,0xc000,0x0100,0x0003,0x20,0x48,0x7b,
  0x40,0x03,0xc000,0x0100,0x0003,0x0f,0xed,0x10,
  0x40,0x03,0xc000,0x0100,0x0003,0x17,0x00,0x34,
  0x40,0x03,0xc000,0x0100,0x0003,0x18,0x00,0x32,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xc0,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xa0,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xb0,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xa8,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xa4,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xa2,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xa1,
  0xc0,0x02,0xc000,0x0019,0x0002,
  0x40,0x03,0xc000,0x0100,0x0003,0x14,0x5a,0xa1,
  0x40,0x03,0xc000,0x0100,0x0003,0x18,0x07,0xff,
  0x40,0x03,0xc000,0x0100,0x0003,0x0f,0x7c,0x82,
  0x40,0x03,0xc000,0x0100,0x0003,0x06,0x3f,0xff,
  0x40,0x03,0xc000,0x0100,0x0003,0x07,0x17,0x1b,
  0x40,0x03,0xc000,0x0100,0x0003,0x08,0x0d,0xff,
  0x40,0x03,0xc000,0x0100,0x0003,0x0d,0x0d,0x80,
  0x40,0x03,0xc000,0x0100,0x0003,0x18,0x07,0xff,
  0x40,0x03,0xc000,0x0100,0x0003,0x17,0x00,0x33,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0x01,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x8c,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x8c,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x4c,0x18,0x6a,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x4d,0x02,0x58,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x4e,0x00,0x0a,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x65,0x00,0x76,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x66,0x02,0x3b,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x67,0x05,0xf3,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x6a,0x0d,0xca,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x6b,0x0a,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x6c,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x6d,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x6e,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x6f,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x70,0x00,0x28,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x71,0xce,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x72,0x48,0x98,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x73,0x58,0x5a,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x4b,0x00,0xb7,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x9b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x9b,0x00,0x04,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0xc0,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x1b,0xff,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x1a,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0x01,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x02,0x0a,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x02,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x86,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0xc0,0x00,
  0xc0,0x02,0xc000,0x0007,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0x01,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0xc0,0x00,
  0x40,0x03,0xc000,0x0100,0x0003,0x1b,0x41,0x12,
  0x40,0x03,0xc000,0x0100,0x0003,0x1a,0x00,0x09,
  0x40,0x03,0x9000,0x0100,0x0004,0x04,0x01,0x01,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x1d,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x1d,0x00,0x23,
  0x40,0x03,0x9000,0x0100,0x0002,0x00,0x1a,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1a,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1c,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1a,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0a,0x00,0xc0,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x01,0x3f,0xfe,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x02,0x0f,0x50,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x03,0x0c,0x38,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x04,0x0c,0x38,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x12,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x12,0x01,0x84,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x11,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x10,0x02,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x11,0x20,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0e,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0f,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x00,0x60,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0c,0x04,0x27,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x60,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x61,0x1f,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5f,0x00,0xd5,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x57,0x72,0x79,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x58,0x68,0x67,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x59,0x52,0x3d,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5a,0x72,0x79,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5b,0x65,0x36,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5c,0x4d,0x65,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5d,0x76,0xdc,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5e,0x65,0x36,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb4,0x04,0x09,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xbb,0x41,0x02,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x5f,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5f,0x00,0xd5,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x52,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x52,0x00,0x61,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb2,0x00,0x20,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xbd,0xe0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xc0,0xe0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xe1,0xe0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0a,0xe0,0x40,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x1f,0xf0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x20,0x1f,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd3,0x1f,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x20,0x66,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x25,0x00,0x65,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x11,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x11,0x69,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd7,0x01,0x48,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd8,0xff,0xf0,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd9,0x00,0xa4,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xda,0xff,0xf8,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x75,0x01,0xc4,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x76,0x01,0x40,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x77,0x00,0x3a,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x78,0xff,0xa5,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x79,0xff,0xcb,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7a,0x00,0x25,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7b,0x00,0x2d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7c,0xff,0xf5,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7d,0xff,0xdd,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7e,0xff,0xfc,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7f,0x00,0x19,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x80,0x00,0x0b,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x81,0xff,0xf2,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x82,0xff,0xf2,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x83,0x00,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x84,0x00,0x0d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x85,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x86,0xff,0xf7,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x87,0xff,0xfb,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x88,0x00,0x04,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x89,0x00,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x8a,0x00,0x03,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x8b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x8c,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xde,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xdf,0x10,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xe4,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x09,0x00,0x1f,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xcd,0x20,0x0f,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x22,0x06,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x23,0x06,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x24,0x06,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xfa,0x0c,0xd5,
  0x40,0x03,0x9000,0x0100,0x0002,0x00,0x00,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x00,0x20,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x06,0x00,0x02,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x07,0x00,0x08,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x08,0x10,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0b,0x00,0x2d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0c,0xc6,0xc0,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0d,0x00,0x2d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0e,0xc6,0xc0,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0f,0x03,0x37,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x10,0xf9,0x80,
  0x40,0x03,0x9000,0x0100,0x0002,0x00,0x00,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x00,0xa0,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x00,0x20,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x05,0x04,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x02,0x3a,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x02,0x3a,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xf9,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xfc,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x02,0x3a,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xe1,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xe7,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xed,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xf3,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xe2,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xe8,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xee,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xf4,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xe3,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xe9,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xef,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0xf5,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1d,0x01,0x15,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1e,0x66,0xab,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x1d,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x1d,0x00,0x23,
  0x40,0x03,0x9000,0x0100,0x0002,0x00,0x1a,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1a,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1c,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x1a,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x0a,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x01,0x3f,0xfe,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x02,0x04,0x53,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x03,0x0f,0x1a,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x04,0x0c,0x08,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x12,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x12,0x01,0xa4,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x11,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x10,0x02,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x11,0x69,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0e,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0f,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x00,0x60,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0c,0x04,0x27,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x60,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x61,0x1f,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5f,0x00,0xd5,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x57,0x72,0x79,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x58,0x68,0x67,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x59,0x52,0x3d,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5a,0x72,0x79,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5b,0x65,0x36,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5c,0x4d,0x65,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5d,0x76,0xdc,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5e,0x65,0x36,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb4,0x04,0x09,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xbb,0x41,0x02,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x5f,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x5f,0x00,0xd5,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x52,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x52,0x00,0x61,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xb2,0x00,0x20,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xbd,0xe0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xc0,0xe0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xe1,0xe0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0a,0xe0,0x40,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x1f,0xf0,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x20,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd3,0x1f,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x20,0x66,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x25,0x00,0x65,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x11,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x11,0x69,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd7,0x01,0x48,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd8,0xff,0xf0,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xd9,0x00,0xa4,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xda,0xff,0xf8,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x75,0x01,0xc4,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x76,0x01,0x40,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x77,0x00,0x3a,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x78,0xff,0xa5,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x79,0xff,0xcb,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7a,0x00,0x25,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7b,0x00,0x2d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7c,0xff,0xf5,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7d,0xff,0xdd,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7e,0xff,0xfc,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x7f,0x00,0x19,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x80,0x00,0x0b,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x81,0xff,0xf2,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x82,0xff,0xf2,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x83,0x00,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x84,0x00,0x0d,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x85,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x86,0xff,0xf7,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x87,0xff,0xfb,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x88,0x00,0x04,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x89,0x00,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x8a,0x00,0x03,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x8b,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x8c,0xff,0xff,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xde,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xdf,0x10,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xe4,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x09,0x00,0x1f,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xcd,0x20,0x0f,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x22,0x06,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x23,0x06,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x24,0x06,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0xfa,0x0c,0xd5,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x02,0x40,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x03,0x02,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x20,0x86,0x80,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x25,0x01,0x08,
  0x40,0x03,0x9000,0x0100,0x0002,0x02,0x38,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x00,0x06,0x02,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x11,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x10,0x02,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x11,0x69,0x06,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0e,0x00,0x01,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x0f,0x00,0x00,
  0x40,0x03,0x9000,0x0100,0x0002,0x01,0x2b,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x2b,0x00,0x22,
  0x40,0x03,0x9000,0x0100,0x0004,0x01,0x2c,0x07,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x06,0x04,0x00,
  0x40,0x03,0x9000,0x0100,0x0004,0x05,0x0b,0x02,0x05,
};

static uint16_t express_chdata[]={
  13,0x4e,0x5a,0xc0,0xe0,0xf0,0xe8,0xe4,0xe6,0xe7,0xe7,
  14,0x4f,0x5a,0xc0,0xe0,0xf0,0xe8,0xe4,0xe2,0xe1,0xe0,
  15,0x50,0x5a,0xc0,0xe0,0xd0,0xd8,0xdc,0xda,0xd9,0xd9,
  16,0x51,0x5a,0xc0,0xe0,0xd0,0xd8,0xd4,0xd2,0xd3,0xd3,
  17,0x52,0x5a,0xc0,0xe0,0xd0,0xc8,0xcc,0xce,0xcd,0xcd,
  18,0x53,0x5a,0xc0,0xe0,0xd0,0xc8,0xc4,0xc6,0xc7,0xc8,
  19,0x54,0x5a,0xc0,0xe0,0xd0,0xc8,0xc4,0xc2,0xc3,0xc2,
  20,0x55,0x5a,0xc0,0xa0,0xb0,0xb8,0xbc,0xbe,0xbd,0xbd,
  21,0x56,0x5a,0xc0,0xa0,0xb0,0xb8,0xb4,0xb6,0xb7,0xb8,
  22,0x57,0x5a,0xc0,0xa0,0xb0,0xb8,0xb4,0xb2,0xb3,0xb3,
  23,0x58,0x5a,0xc0,0xa0,0xb0,0xa8,0xac,0xae,0xad,0xae,
  24,0x59,0x5a,0xc0,0xa0,0xb0,0xa8,0xac,0xaa,0xa9,0xa9,
  25,0x5a,0x5a,0xc0,0xa0,0xb0,0xa8,0xa4,0xa6,0xa5,0xa5,
  26,0x5b,0x5a,0xc0,0xa0,0xb0,0xa8,0xa4,0xa2,0xa1,0xa0,
  27,0x5c,0x5a,0xc0,0xa0,0x90,0x98,0x9c,0x9e,0x9d,0x9c,
  28,0x5d,0x5a,0xc0,0xa0,0x90,0x98,0x9c,0x9a,0x99,0x98,
  29,0x5e,0x5a,0xc0,0xa0,0x90,0x98,0x94,0x96,0x95,0x94,
  30,0x5f,0x50,0xc0,0xe0,0xf0,0xe8,0xec,0xea,0xeb,0xec,
  31,0x60,0x50,0xc0,0xe0,0xf0,0xe8,0xe4,0xe6,0xe7,0xe6,
  32,0x61,0x50,0xc0,0xe0,0xf0,0xe8,0xe4,0xe2,0xe1,0xe1,
  33,0x62,0x50,0xc0,0xe0,0xd0,0xd8,0xdc,0xda,0xdb,0xdb,
  34,0x63,0x50,0xc0,0xe0,0xd0,0xd8,0xd4,0xd6,0xd5,0xd6,
  35,0x64,0x50,0xc0,0xe0,0xd0,0xd8,0xd4,0xd2,0xd1,0xd0,
  36,0x65,0x50,0xc0,0xe0,0xd0,0xc8,0xcc,0xce,0xcd,0xcc,
  37,0x66,0x50,0xc0,0xe0,0xd0,0xc8,0xc4,0xc6,0xc7,0xc7,
  38,0x67,0x50,0xc0,0xe0,0xd0,0xc8,0xc4,0xc2,0xc3,0xc3,
  39,0x68,0x50,0xc0,0xa0,0xb0,0xb8,0xbc,0xbe,0xbf,0xbf,
  40,0x69,0x50,0xc0,0xa0,0xb0,0xb8,0xbc,0xba,0xbb,0xba,
  41,0x6a,0x50,0xc0,0xa0,0xb0,0xb8,0xb4,0xb6,0xb7,0xb6,
  42,0x6b,0x50,0xc0,0xa0,0xb0,0xb8,0xb4,0xb2,0xb1,0xb2,
  43,0x6c,0x50,0xc0,0xa0,0xb0,0xa8,0xac,0xae,0xad,0xae,
  44,0x6d,0x50,0xc0,0xa0,0xb0,0xa8,0xac,0xaa,0xa9,0xaa,
  45,0x6e,0x50,0xc0,0xa0,0xb0,0xa8,0xa4,0xa6,0xa7,0xa6,
  46,0x6f,0x50,0xc0,0xa0,0xb0,0xa8,0xa4,0xa2,0xa3,0xa3,
  47,0x70,0x50,0xc0,0xa0,0x90,0x98,0x9c,0x9e,0x9f,0x9f,
  48,0x71,0x50,0xc0,0xa0,0x90,0x98,0x9c,0x9a,0x9b,0x9c,
  49,0x72,0x50,0xc0,0xa0,0x90,0x98,0x9c,0x9a,0x99,0x98,
  50,0x73,0x50,0xc0,0xa0,0x90,0x98,0x94,0x96,0x95,0x94,
  51,0x74,0x58,0xc0,0xe0,0xf0,0xe8,0xec,0xee,0xed,0xee,
  52,0x75,0x58,0xc0,0xe0,0xf0,0xe8,0xec,0xea,0xe9,0xe9,
  53,0x76,0x58,0xc0,0xe0,0xf0,0xe8,0xe4,0xe6,0xe5,0xe5,
  54,0x77,0x58,0xc0,0xe0,0xf0,0xe8,0xe4,0xe2,0xe1,0xe0,
  55,0x78,0x58,0xc0,0xe0,0xd0,0xd8,0xdc,0xda,0xdb,0xdc,
  56,0x79,0x58,0xc0,0xe0,0xd0,0xd8,0xd4,0xd6,0xd7,0xd7,
  57,0x7a,0x58,0xc0,0xe0,0xd0,0xd8,0xd4,0xd2,0xd3,0xd3,
  58,0x7b,0x58,0xc0,0xe0,0xd0,0xc8,0xcc,0xce,0xcf,0xcf,
  59,0x7c,0x58,0xc0,0xe0,0xd0,0xc8,0xcc,0xca,0xcb,0xcc,
  60,0x7d,0x58,0xc0,0xe0,0xd0,0xc8,0xc4,0xc6,0xc7,0xc8,
  61,0x7e,0x58,0xc0,0xe0,0xd0,0xc8,0xc4,0xc6,0xc5,0xc4,
  62,0x7f,0x58,0xc0,0xe0,0xd0,0xc8,0xc4,0xc2,0xc1,0xc1,
};

#define DATALEN(a) (sizeof(a) / sizeof(uint16_t))

static int express_set_freq(struct jdvbt90502_state *state, u32 freq)
{
	u32 freqfix;
	int channel;
	int rval;
	int index,i;
	size_t length;
	deb_fe("%s: entered.\n", __func__);

	freqfix=freq-3000000;
	channel=freqfix/6000000;

	channel-=78;
	if(channel<0 || channel>49)return -EREMOTEIO;
	index=channel*11;
	
	deb_fe("express channel set to %d\n",channel+13);
	express_usbdata[1887]=express_chdata[index+1];
	for(i=0;i<8;i++){
		express_usbdata[1942+(i*13)]=express_chdata[index+2];
		express_usbdata[1943+(i*13)]=express_chdata[index+3+i];
	}
	length=(sizeof(express_usbdata)/sizeof(uint16_t));
	rval=usb_ctrl_sends(state,express_usbdata,length,NULL,0);
	return 0;
}
static int express_read_status(struct dvb_frontend *fe, fe_status_t *state){
  deb_fe("%s: entered.\n", __func__);
/*  *state = FE_HAS_SIGNAL | FE_HAS_CARRIER |
    FE_HAS_VITERBI | FE_HAS_SYNC | FE_HAS_LOCK;
  return 0; */
	u8 result;
	int ret;

	*state = 0;

	ret = jdvbt90502_reg_read(fe->demodulator_priv, 0x96, &result, 1);
	if (ret)
		goto error;
	if (result & 0xE0) {
		deb_fe("%s: RLOCK detected.\n", __func__);
		*state = FE_HAS_SIGNAL | FE_HAS_CARRIER |
			 FE_HAS_VITERBI | FE_HAS_SYNC | FE_HAS_LOCK;
		return 0;
	}

	ret = jdvbt90502_reg_read(fe->demodulator_priv, 0x80, &result, 1);
	deb_fe("%s: r0x80 == 0x%02hhx\n", __func__, result);
	if (ret)
		goto error;
	if (result & 0xC0)
		return 0;
	*state = FE_HAS_SIGNAL | FE_HAS_CARRIER;

	if (result & 0x04)
		return 0;
	*state |= FE_HAS_VITERBI;

	ret = jdvbt90502_reg_read(fe->demodulator_priv, 0xB0, &result, 1);
	if (ret)
		goto error;
	if ((result & 0x0F) >= 8)
		*state |= FE_HAS_SYNC | FE_HAS_LOCK;

	return 0;

error:
	deb_fe("%s:ret == %d\n", __func__, ret);
	return -EREMOTEIO;

}

static int jdvbt90502_read_status(struct dvb_frontend *fe, fe_status_t *state)
{
	u8 result;
	int ret;

	*state = 0;

	ret = jdvbt90502_reg_read(fe->demodulator_priv, 0x96, &result, 1);
	if (ret)
		goto error;
	if (result & 0xE0) {
		deb_fe("%s: RLOCK detected.\n", __func__);
		*state = FE_HAS_SIGNAL | FE_HAS_CARRIER |
			 FE_HAS_VITERBI | FE_HAS_SYNC | FE_HAS_LOCK;
		return 0;
	}

	ret = jdvbt90502_reg_read(fe->demodulator_priv, 0x80, &result, 1);
	deb_fe("%s: r0x80 == 0x%02hhx\n", __func__, result);
	if (ret)
		goto error;
	if (result & 0xC0)
		return 0;
	*state = FE_HAS_SIGNAL | FE_HAS_CARRIER;

	if (result & 0x04)
		return 0;
	*state |= FE_HAS_VITERBI;

	ret = jdvbt90502_reg_read(fe->demodulator_priv, 0xB0, &result, 1);
	if (ret)
		goto error;
	if ((result & 0x0F) >= 8)
		*state |= FE_HAS_SYNC | FE_HAS_LOCK;

	return 0;

error:
	deb_fe("%s:ret == %d\n", __func__, ret);
	return -EREMOTEIO;
}

static int jdvbs90502_read_status(struct dvb_frontend *fe, fe_status_t *state)
{
	u8 res1, res2, res3;
	int ret;

	*state = 0;

	ret = jdvbs90502_reg_read(fe->demodulator_priv, 0xC3, &res1, 1);
	if (ret)
		goto error;
	if (res1 & 0x80)
		return 0;
	*state |= FE_HAS_SIGNAL;

	ret = jdvbs90502_reg_read(fe->demodulator_priv, 0xBA, &res2, 1);
	if (ret)
		goto error;
	if (res2 & 0x80)
		return 0;
	*state |= FE_HAS_CARRIER;
	if (res1 & 0x40)
		return 0;

	ret = jdvbs90502_reg_read(fe->demodulator_priv, 0xC5, &res3, 1);
	if (ret)
		goto error;
	if (!(res3 & 0x03))
		return 0;
	*state |= FE_HAS_SYNC | FE_HAS_VITERBI | FE_HAS_LOCK;

	return 0;

error:
	deb_fe("%s:ret == %d\n", __func__, ret);
	return -EREMOTEIO;
}

static int jdvbt90502_read_signal_strength(struct dvb_frontend *fe,
					   u16 *strength)
{
	int ret;
	u8 rbuf[3];
	u32 val;

	*strength = 0;

	ret = jdvbt90502_reg_read(fe->demodulator_priv, 0x8B,
				  rbuf, sizeof(rbuf));

	if (ret) {
		deb_fe("%s:ret == %d\n", __func__, ret);
		return -EREMOTEIO;
	}

	val = (rbuf[0] << 16) | (rbuf[1] << 8) | rbuf[2];
	deb_fe("%s:demod_sig_condist == 0x%0x\n", __func__, val);
	/*
	 * val: average constellation distribution?
	 * signal_strength := 68.4344993 - 10.0 * log10(val) +
	 *                        pow(log10(val) - 5.045, 4.0) / 4.09;
	 */
	*strength = (0xffffff - val) >> 8;

	return 0;
}

static int jdvbs90502_read_signal_strength(struct dvb_frontend *fe,
					   u16 *strength)
{
	int ret;
	u8 rbuf[2];
	u16 val;

	*strength = 0;

	ret = jdvbs90502_reg_read(fe->demodulator_priv, 0xBC,
				  rbuf, sizeof(rbuf));

	if (ret) {
		deb_fe("%s:ret == %d\n", __func__, ret);
		return -EREMOTEIO;
	}

	val = (rbuf[0] << 8) | rbuf[1];
	deb_fe("%s:demod_sig_condist == 0x%04hx\n", __func__, val);
	/*
	 * val: average constellation distribution?
	 * C/N[db] := -1.6346 * P^5 + 14.341 * P^4 - 50.259 * P^3 + 88.977 * P^2
	 *              - 89.565 * P + 58.857
	 * (P:= sqrt(val - 3000) / 64 (val >= 3000)
	 * or C/N[db] := -0.14 * sqrt(val - 3000) + 28.7
	 */
	*strength = 0xffff - val;
	return 0;
}

static uint16_t express_usbsignal1[] = {
  0x40,0x03,0x9000,0x0100,0x0002,0x02,0x1e,
  0xc0,0x02,0x9000,0x0100,0x0002,
};

static uint16_t express_usbsignal2[] = {
  0x40,0x03,0x9000,0x0100,0x0002,0x02,0x1f,
  0xc0,0x02,0x9000,0x0100,0x0002
};

static int express_read_signal_strength(struct dvb_frontend *fe,
					   u16 *strength)
{
  struct jdvbt90502_state *state = fe->demodulator_priv;
  size_t length;
  uint8_t buf1e[2],buf1f[2];
  deb_fe("%s: entered.\n", __func__);
  length=(sizeof(express_usbsignal1)/sizeof(uint16_t));
  usb_ctrl_sends(state,express_usbsignal1,length,buf1e,2);
  length=(sizeof(express_usbsignal2)/sizeof(uint16_t));
  usb_ctrl_sends(state,express_usbsignal2,length,buf1f,2);
  if(buf1f[1] <=36){
    *strength=0;
  }else{
    *strength=20;
  }
  return 0;
}

/* filter out un-supported properties to notify users */
static int jdvbt90502_set_property(struct dvb_frontend *fe,
				   struct dtv_property *tvp)
{
	int r = 0;

	switch (tvp->cmd) {
	case DTV_DELIVERY_SYSTEM:
		if (tvp->u.data != SYS_ISDBT)
			r = -EINVAL;
		break;
	case DTV_CLEAR:
	case DTV_TUNE:
	case DTV_FREQUENCY:
		break;
	default:
		r = -EINVAL;
	}
	return r;
}

static int jdvbs90502_set_property(struct dvb_frontend *fe,
				   struct dtv_property *tvp)
{
	int r = 0;

	switch (tvp->cmd) {
	case DTV_DELIVERY_SYSTEM:
		if (tvp->u.data != SYS_ISDBS)
			r = -EINVAL;
		break;
	case DTV_CLEAR:
	case DTV_TUNE:
	case DTV_FREQUENCY:
	case DTV_ISDBS_TS_ID:
		break;
	default:
		r = -EINVAL;
	}
	return r;
}

static int jdvbt90502_get_frontend(struct dvb_frontend *fe,
				   struct dvb_frontend_parameters *p)
{
	p->inversion = INVERSION_AUTO;
	p->u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
	p->u.ofdm.code_rate_HP = FEC_AUTO;
	p->u.ofdm.code_rate_LP = FEC_AUTO;
	p->u.ofdm.constellation = QAM_64;
	p->u.ofdm.transmission_mode = TRANSMISSION_MODE_AUTO;
	p->u.ofdm.guard_interval = GUARD_INTERVAL_AUTO;
	p->u.ofdm.hierarchy_information = HIERARCHY_AUTO;
	return 0;
}


static int jdvbt90502_set_frontend(struct dvb_frontend *fe,
				   struct dvb_frontend_parameters *p)
{
	/**
	 * NOTE: ignore all the paramters except frequency.
	 *       others should be fixed to the proper value for ISDB-T,
	 *       but don't check here.
	 */

	struct jdvbt90502_state *state = fe->demodulator_priv;
	int ret;

	deb_fe("%s: Freq:%d\n", __func__, fe->dtv_property_cache.frequency);

	/* for recovery from DTV_CLEAN */
	fe->dtv_property_cache.delivery_system = SYS_ISDBT;

	ret = jdvbt90502_set_freq(state, fe->dtv_property_cache.frequency);
	if (ret) {
		deb_fe("%s:ret == %d\n", __func__, ret);
		return -EREMOTEIO;
	}

	return 0;
}

static int jdvbs90502_set_frontend(struct dvb_frontend *fe,
				   struct dvb_frontend_parameters *p)
{
	struct jdvbt90502_state *state = fe->demodulator_priv;
	int ret;

	deb_fe("%s: Freq:%d\n", __func__, fe->dtv_property_cache.frequency);

	/* for recovery from DTV_CLEAN */
	fe->dtv_property_cache.delivery_system = SYS_ISDBS;

	ret = jdvbs90502_set_freq(state,
				fe->dtv_property_cache.frequency,
				fe->dtv_property_cache.isdbs_ts_id);
	if (ret) {
		deb_fe("%s:ret == %d\n", __func__, ret);
		return -EREMOTEIO;
	}
	return 0;
}

static int express_set_frontend(struct dvb_frontend *fe,
				   struct dvb_frontend_parameters *p)
{
  /**
   * NOTE: ignore all the paramters except frequency.
   *       others should be fixed to the proper value for ISDB-T,
   *       but don't check here.
   */

  struct jdvbt90502_state *state = fe->demodulator_priv;
  int ret;

  deb_fe("%s: Freq:%d\n", __func__, fe->dtv_property_cache.frequency);

  /* for recovery from DTV_CLEAN */
  fe->dtv_property_cache.delivery_system = SYS_ISDBT;

  ret = express_set_freq(state, fe->dtv_property_cache.frequency);
  if (ret) {
    deb_fe("%s:ret == %d\n", __func__, ret);
    return -EREMOTEIO;
  }

  return 0;
}

static int jdvbt90502_get_tune_settings(struct dvb_frontend *fe,
				   struct dvb_frontend_tune_settings *s)
{
	/* default min_delay for OFDM is just too short for Friio */
	s->min_delay_ms = 400;
	s->step_size = fe->ops.info.frequency_stepsize * 2;
	s->max_drift = fe->ops.info.frequency_stepsize * 2 + 1;
	return 0;
}

static int jdvbs90502_get_tune_settings(struct dvb_frontend *fe,
				   struct dvb_frontend_tune_settings *s)
{
	/* default min_delay (for QPSK) is just too short for Friio */
	s->min_delay_ms = 400;
	s->step_size = 0;	/* no zigzag */
	s->max_drift = 0;
	return 0;
}

static int jdvbt90502_sleep(struct dvb_frontend *fe)
{
	static u8 sleep_code[2] = {0x03, 0x10};
	int ret;
	struct i2c_msg msg;
	struct jdvbt90502_state *state = fe->demodulator_priv;

	deb_fe("%s called.\n", __func__);

	msg.addr = state->config.demod_address;
	msg.flags = 0;
	msg.len = 2;
	msg.buf = sleep_code;

	ret = i2c_transfer(state->i2c, &msg, 1);
	if (ret != 1)
		goto error;
	return 0;

error:
	deb_fe("%s: sleep command failed. ret==%d\n", __func__, ret);
	return -EREMOTEIO;
}

static int ce5037_sleep(struct dvb_frontend *fe)
{
	int ret;
	struct i2c_msg msg;
	struct jdvbt90502_state *state = fe->demodulator_priv;
	u8 tuner_sleep_cmd[] = {
		JDVBT90502_2ND_I2C_REG,
		state->config.bk_pll_address << 1, 0x0F, 0x91
	};

	msg.addr = state->config.bk_demod_address;
	msg.flags = 0;
	msg.len = sizeof(tuner_sleep_cmd);
	msg.buf = tuner_sleep_cmd;
	ret = i2c_transfer(state->i2c, &msg, 1);
	if (ret != 1)
		goto error;
	return 0;
error:
	deb_fe("%s: sleep command failed. ret==%d\n", __func__, ret);
	return -EREMOTEIO;
}

static int jdvbs90502_sleep(struct dvb_frontend *fe)
{
	jdvbs90502_single_reg_write(fe->demodulator_priv, 0x17, 0x01);
	return jdvbt90502_sleep(fe);
}

static int express_sleep(struct dvb_frontend *fe)
{
  deb_fe("%s: entered.\n", __func__);
  return 0;
}
/**
 * (reg, val) commad list to initialize this module.
 *  captured on a Windows box.
 */
static u8 init_code[][2] = {
	{0x01, 0x40},
	{0x03, 0x00},
	{0x04, 0x38},
	{0x05, 0x40},
	{0x07, 0x40},
	{0x0F, 0x4F},
	{0x11, 0x21},
	{0x12, 0x0B},
	{0x13, 0x2F},
	{0x14, 0x31},
	{0x16, 0x02},
	{0x21, 0xC4},
	{0x22, 0x20},
	{0x2C, 0x79},
	{0x2D, 0x34},
	{0x2F, 0x00},
	{0x30, 0x28},
	{0x31, 0x31},
	{0x32, 0xDF},
	{0x38, 0x01},
	{0x39, 0x78},
	{0x3B, 0x33},
	{0x3C, 0x33},
	{0x48, 0x90},
	{0x51, 0x68},
	{0x5E, 0x38},
	{0x71, 0x00},
	{0x72, 0x08},
	{0x77, 0x00},
	{0xC0, 0x21},
	{0xC1, 0x10},
	{0xE4, 0x1A},
	{0xEA, 0x1F},
	{0x77, 0x00},
	{0x71, 0x00},
	{0x71, 0x00},
	{0x76, 0x0C},
};

static const int init_code_len = sizeof(init_code) / sizeof(u8[2]);

static int jdvbt90502_init(struct dvb_frontend *fe)
{
	int i = -1;
	int ret;
	struct i2c_msg msg;

	struct jdvbt90502_state *state = fe->demodulator_priv;

	deb_fe("%s called.\n", __func__);

	msg.addr = state->config.demod_address;
	msg.flags = 0;
	msg.len = 2;
	for (i = 0; i < init_code_len; i++) {
		msg.buf = init_code[i];
		ret = i2c_transfer(state->i2c, &msg, 1);
		if (ret != 1)
			goto error;
	}
	fe->dtv_property_cache.delivery_system = SYS_ISDBT;
	msleep(50);

	return 0;

error:
	deb_fe("%s: init_code[%d] failed. ret==%d\n", __func__, i, ret);
	return -EREMOTEIO;
}


static int ce5037_init(struct dvb_frontend *fe)
{
	int ret;
	struct i2c_msg msg;
	struct jdvbt90502_state *state = fe->demodulator_priv;
	u8 result;

	u8 tuner_init_cmd0[] = {
		JDVBT90502_2ND_I2C_REG,
		state->config.bk_pll_address << 1, 0x00,
		0x08, 0x32, 0x02, 0x40, 0x6C, 0x47, 0x02, 0x30,
		0x20, 0xA2, 0xF1, 0x38, 0xD0, 0x10, 0xB0, 0x01
	};
	u8 tuner_init_cmd1[] = {
		JDVBT90502_2ND_I2C_REG,
		state->config.bk_pll_address << 1, 0x06, 0x62
	};

	msg.addr = state->config.bk_demod_address;
	msg.flags = 0;
	msg.len = sizeof(tuner_init_cmd0);
	msg.buf = tuner_init_cmd0;
	ret = i2c_transfer(state->i2c, &msg, 1);
	if (ret != 1)
		goto error;

	msleep(1);
	result = 0x07;
	ret = jdvbs90502_pll_read(state, &result);
	if (ret != 0)
		goto error;
	if (!(result & 0x80)) {
		deb_fe("%s: baseband lock failed.\n", __func__);
		return -EREMOTEIO;
	}

	msg.len = sizeof(tuner_init_cmd1);
	msg.buf = tuner_init_cmd1;
	ret = i2c_transfer(state->i2c, &msg, 1);
	if (ret != 1)
		goto error;
	return 0;

error:
	deb_fe("%s: init command failed. ret==%d\n", __func__, ret);
	return -EREMOTEIO;
}
static int express_init(struct dvb_frontend *fe)
{
  deb_fe("%s: entered.\n", __func__);
  return 0;
}
static u8 init_bk_ofdm[][2] = {
	{0x01, 0x40},
	{0x03, 0x00},	/* wakeup (both 8PSK and) OFDM */
	{0x04, 0x30},
	{0x05, 0x40},
	{0x07, 0x40},
	{0x0F, 0xFF},
	{0x11, 0x27},
	{0x12, 0x02},	/* unnecessary? */
	{0x13, 0x2F},
	{0x14, 0x31},
	{0x16, 0x02},
	{0x22, 0x20},
	{0x24, 0x70},
	{0x2C, 0x79},
	{0x2D, 0x34},
	{0x30, 0x28},
	{0x31, 0x31},
	{0x32, 0xDF},	/* unnecessary? */
	{0x38, 0x01},
	{0x39, 0x78},
	{0x48, 0x90},
	{0x51, 0x68},
	{0x5E, 0x38},
	{0xC0, 0x21},
	{0xC1, 0x10},
	{0xE4, 0x1A},
	{0xEA, 0x1F},
	{0x77, 0x00},
	{0x71, 0x00},
	{0x71, 0x00},

	{0x0, 0x0}
};


static u8 init_bk_8psk[][2] = {
	{0x01, 0x10},
	{0x02, 0x00},
	{0x03, 0x00},
	{0x08, 0x00},
	{0x0C, 0x59},
	{0x0D, 0xF2},
	{0x0F, 0x51},	/* ?? 0x50? */
	{0x10, 0xB2},
	{0x1A, 0x04},
	{0x1B, 0x04},
	{0x38, 0x40},
	{0x39, 0x10},
	{0x3C, 0x08},
	{0x3D, 0x96},
	{0x3E, 0xD9},
	{0x3F, 0xB6},
	{0x40, 0x5D},
	{0x41, 0x99},
	{0x44, 0x58},
	{0x45, 0x54},
	{0x46, 0x30},
	{0x47, 0x41},
	{0x49, 0x16},
	{0x4A, 0x41},
	{0x4B, 0x41},
	{0x4F, 0x5E},
	{0x50, 0x04},
	{0x51, 0xC0},
	{0x52, 0x8A},
	{0x53, 0x13},
	{0x54, 0x90},
	{0x55, 0x92},
	{0x58, 0x6D},
	{0x59, 0x6F},
	{0x5A, 0x2E},
	{0x5B, 0x23},
	{0x5D, 0x33},
	{0x87, 0x04},
	{0x8D, 0x00},
	{0x8F, 0x40},
	{0x90, 0x10},
	{0xA5, 0x07},
	{0xA3, 0xF7},	/* 0xFF? */

	{0x0, 0x0}
};

static int jdvbs90502_init(struct dvb_frontend *fe)
{
	int i = -1;
	int ret;
	struct i2c_msg msg;
	struct jdvbt90502_state *state = fe->demodulator_priv;

	deb_fe("%s called.\n", __func__);

	/* system reset, no error check, as it may fail to get ACK'ed */
	jdvbs90502_single_reg_write(state, 0x01, 0x80);

	msg.addr = state->config.demod_address;
	msg.flags = 0;
	msg.len = 2;
	for (i = 0; init_bk_ofdm[i][0] != 0; i++) {
		msg.buf = init_bk_ofdm[i];
		ret = i2c_transfer(state->i2c, &msg, 1);
		if (ret != 1)
			goto error;
	}

	msg.addr = state->config.bk_demod_address;
	for (i = 0; init_bk_8psk[i][0] != 0; i++) {
		msg.buf = init_bk_8psk[i];
		ret = i2c_transfer(state->i2c, &msg, 1);
		if (ret != 1)
			goto error;
	}

	fe->dtv_property_cache.delivery_system = SYS_ISDBS;
	return 0;

error:
	deb_fe("%s: init_code[%d] failed. ret==%d\n", __func__, i, ret);
	return -EREMOTEIO;
}


static void jdvbt90502_release(struct dvb_frontend *fe)
{
	struct jdvbt90502_state *state = fe->demodulator_priv;
	kfree(state);
}
static uint16_t express_usbend[]={
  0x40,0x01,0x0002,0x0011,0x0000,
  0x40,0x01,0x0000,0x0011,0x0000,
  0x40,0x03,0x1200,0x0100,0x0002,0x03,0x80,
  0xc0,0x02,0x1200,0x0100,0x0002,
  0x40,0x03,0x9000,0x0100,0x0002,0x03,0x80,
  0xc0,0x02,0x9000,0x0100,0x0002,
  0x40,0x01,0x0004,0x0030,0x0000,
  0x40,0x01,0x0019,0x0000,0x0000,
};

static void express_release(struct dvb_frontend *fe){
  struct jdvbt90502_state *state = fe->demodulator_priv;
  size_t length;
  deb_fe("%s: entered.\n", __func__);

  length=(sizeof(express_usbend)/sizeof(uint16_t));
  usb_ctrl_sends(state,express_usbend,length,NULL,0);
  kfree(state);
}  

static struct dvb_frontend_ops jdvbt90502_ops = {

	.info = {
		.name			= "Comtech JDVBT90502 ISDB-T",
		.type			= FE_OFDM,
		.frequency_min		= 473000000, /* UHF 13ch, center */
		.frequency_max		= 767142857, /* UHF 62ch, center */
		.frequency_stepsize	= JDVBT90502_PLL_CLK /
							JDVBT90502_PLL_DIVIDER,

		/* NOTE: this driver ignores all parameters but frequency. */
		.caps = FE_CAN_INVERSION_AUTO |
			FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 | FE_CAN_FEC_3_4 |
			FE_CAN_FEC_4_5 | FE_CAN_FEC_5_6 | FE_CAN_FEC_6_7 |
			FE_CAN_FEC_7_8 | FE_CAN_FEC_8_9 | FE_CAN_FEC_AUTO |
			FE_CAN_QAM_16 | FE_CAN_QAM_64 | FE_CAN_QAM_AUTO |
			FE_CAN_TRANSMISSION_MODE_AUTO |
			FE_CAN_GUARD_INTERVAL_AUTO |
			FE_CAN_HIERARCHY_AUTO,
	},

	.release = jdvbt90502_release,

	.init = jdvbt90502_init,
	.sleep = jdvbt90502_sleep,

	.set_property = jdvbt90502_set_property,

	.set_frontend = jdvbt90502_set_frontend,
	.get_tune_settings = jdvbt90502_get_tune_settings,
	.get_frontend = jdvbt90502_get_frontend,

	.read_status = jdvbt90502_read_status,
	.read_signal_strength = jdvbt90502_read_signal_strength,
};

static struct dvb_frontend_ops express_ops = {
  .info = {
    .name                   = "Friio express ISDB-T",
    .type                   = FE_OFDM,
    .frequency_min          = 473000000, /* UHF 13ch, center */
    .frequency_max          = 767142857, /* UHF 62ch, center */
                .frequency_stepsize     = JDVBT90502_PLL_CLK /
    JDVBT90502_PLL_DIVIDER,

    /* NOTE: this driver ignores all parameters but frequency. */
                .caps = FE_CAN_INVERSION_AUTO |
                        FE_CAN_FEC_1_2 | FE_CAN_FEC_2_3 | FE_CAN_FEC_3_4 |
                        FE_CAN_FEC_4_5 | FE_CAN_FEC_5_6 | FE_CAN_FEC_6_7 |
                        FE_CAN_FEC_7_8 | FE_CAN_FEC_8_9 | FE_CAN_FEC_AUTO |
                        FE_CAN_QAM_16 | FE_CAN_QAM_64 | FE_CAN_QAM_AUTO |
                        FE_CAN_TRANSMISSION_MODE_AUTO |
                        FE_CAN_GUARD_INTERVAL_AUTO |
    FE_CAN_HIERARCHY_AUTO,
  },

  .release = express_release,

  .init = express_init,
  .sleep = express_sleep,

  .set_property = jdvbt90502_set_property,

  .set_frontend = express_set_frontend,
  .get_tune_settings = jdvbt90502_get_tune_settings,
  .get_frontend = jdvbt90502_get_frontend,

  .read_status = express_read_status,
  .read_signal_strength = express_read_signal_strength,
};

static struct dvb_frontend_ops jdvbs90502_ops = {

	.info = {
		.name			= "Comtech JDVBS90502 ISDB-S",
		.type			= FE_QPSK,
		.frequency_min		= 950000,
		.frequency_max		= 2150000,
	},

	.release = jdvbt90502_release,

	.init = jdvbs90502_init,
	.sleep = jdvbs90502_sleep,

	.set_property = jdvbs90502_set_property,

	.set_frontend = jdvbs90502_set_frontend,
	.get_tune_settings = jdvbs90502_get_tune_settings,

	.read_status = jdvbs90502_read_status,
	.read_signal_strength = jdvbs90502_read_signal_strength,

	.tuner_ops = {
		.info = {
			.name		= "IBM CE5037",
		},
		.init = ce5037_init,
		.sleep = ce5037_sleep,
	},
};


struct dvb_frontend *jdvbt90502_attach(struct dvb_usb_device *d, int is_bs, int is_express)
{
	struct jdvbt90502_state *state = NULL;
	struct dvb_frontend_ops *ops;

	deb_info("%s called.\n", __func__);

	/* allocate memory for the internal state */
	state = kzalloc(sizeof(struct jdvbt90502_state), GFP_KERNEL);
	if (state == NULL)
		goto error;

	/* setup the state */
	state->i2c = &d->i2c_adap;
	state->udev = d->udev;
	memcpy(&state->config, &friio_fe_config, sizeof(friio_fe_config)); 

	/* create dvb_frontend */
	ops = (is_express) ? &express_ops : &jdvbt90502_ops;
	ops = (is_bs) ? &jdvbs90502_ops : ops;
	memcpy(&state->frontend.ops, ops, sizeof(*ops));
	state->frontend.demodulator_priv = state;

	if (jdvbt90502_init(&state->frontend) < 0)
		goto error;

	return &state->frontend;

error:
	kfree(state);
	return NULL;
}
